@@ -26,11 +26,12 @@ module Distribution.Client.Dependency.TopDown.Constraints (
 
 import Distribution.Client.Dependency.TopDown.Types
 import qualified Distribution.Client.PackageIndex as PackageIndex
-import Distribution.Client.PackageIndex (PackageIndex)
+import Distribution.Client.PackageIndex
+        ( PackageIndex )
 import Distribution.Package
          ( PackageName, PackageId, PackageIdentifier(..)
          , Package(packageId), packageName, packageVersion
-         , Dependency, PackageFixedDeps(depends) )
+         , Dependency )
 import Distribution.Version
          ( Version )
 import Distribution.Client.Utils
@@ -224,14 +225,12 @@ transitionsTo constraints @(Constraints _ available  excluded  _ _)
       SourceOnly           b -> SourceOnly    (g b)
       InstalledAndSource a b -> InstalledAndSource (f a) (g b)
 
-
 -- | We construct 'Constraints' with an initial 'PackageIndex' of all the
 -- packages available.
 --
-empty :: (PackageFixedDeps installed, Package source)
-      => PackageIndex installed
-      -> PackageIndex source
-      -> Constraints installed source reason
+empty :: PackageIndex InstalledPackageEx
+      -> PackageIndex UnconfiguredPackage
+      -> Constraints InstalledPackageEx UnconfiguredPackage reason
 empty installed source =
     Constraints targets pkgs excluded pairs pkgs
   where
@@ -253,27 +252,24 @@ empty installed source =
       , let name   = packageName pkg1
             pkgid1 = packageId pkg1
             pkgid2 = packageId pkg2
-      ,    any ((pkgid1==) . packageId) (depends pkg2)
-        || any ((pkgid2==) . packageId) (depends pkg1) ]
+      ,    any ((pkgid1==) . packageId) (sourceDeps pkg2)
+        || any ((pkgid2==) . packageId) (sourceDeps pkg1) ]
 
 
 -- | The package targets.
 --
-packages :: (Package installed, Package source)
-         => Constraints installed source reason
+packages :: Constraints installed source reason
          -> Set PackageName
 packages (Constraints ts _ _ _ _) = ts
 
 
 -- | The package choices that are still available.
 --
-choices :: (Package installed, Package source)
-        => Constraints installed source reason
+choices :: Constraints installed source reason
         -> PackageIndex (InstalledOrSource installed source)
 choices (Constraints _ available _ _ _) = available
 
-isPaired :: (Package installed, Package source)
-         => Constraints installed source reason
+isPaired :: Constraints installed source reason
          -> PackageId -> Maybe PackageId
 isPaired (Constraints _ _ _ pairs _) (PackageIdentifier name version) =
   case Map.lookup name pairs of
