@@ -18,6 +18,8 @@ module Distribution.Client.Init (
 
     -- * Commands
     initCabal
+  , pvpize
+  , incVersion
 
   ) where
 
@@ -41,10 +43,10 @@ import Data.Function
   ( on )
 import qualified Data.Map as M
 #if !MIN_VERSION_base(4,8,0)
-import Data.Traversable
-  ( traverse )
 import Control.Applicative
   ( (<$>) )
+import Data.Traversable
+  ( traverse )
 #endif
 import Control.Monad
   ( when, unless, (>=>), join, forM_ )
@@ -93,23 +95,38 @@ import Distribution.Simple.PackageIndex
 import Distribution.Text
   ( display, Text(..) )
 
+import Distribution.Client.PackageIndex
+  ( elemByPackageName )
+import Distribution.Client.IndexUtils
+  ( getSourcePackages )
+import Distribution.Client.Types
+  ( SourcePackageDb(..) )
+import Distribution.Client.Setup
+  ( RepoContext(..) )
+
 initCabal :: Verbosity
           -> PackageDBStack
+          -> RepoContext
           -> Compiler
           -> ProgramConfiguration
           -> InitFlags
           -> IO ()
-initCabal verbosity packageDBs comp conf initFlags = do
+initCabal verbosity packageDBs repoCtxt comp conf initFlags = do
 
   installedPkgIndex <- getInstalledPackages verbosity comp packageDBs conf
+  sourcePkgDb <- getSourcePackages verbosity repoCtxt
 
   hSetBuffering stdout NoBuffering
 
-  initFlags' <- extendFlags installedPkgIndex initFlags
+  initFlags' <- extendFlags installedPkgIndex sourcePkgDb initFlags
 
-  writeLicense initFlags'
+  case license initFlags' of
+    Flag PublicDomain -> return ()
+    _                 -> writeLicense initFlags'
   writeSetupFile initFlags'
+  writeChangeLog initFlags'
   createSourceDirectories initFlags'
+  createMainHs initFlags'
   success <- writeCabalFile initFlags'
 
   when success $ generateWarnings initFlags'
@@ -120,9 +137,9 @@ initCabal verbosity packageDBs comp conf initFlags = do
 
 -- | Fill in more details by guessing, discovering, or prompting the
 --   user.
-extendFlags :: InstalledPackageIndex -> InitFlags -> IO InitFlags
-extendFlags pkgIx =
-      getPackageName
+extendFlags :: InstalledPackageIndex -> SourcePackageDb -> InitFlags -> IO InitFlags
+extendFlags pkgIx sourcePkgDb =
+      getPackageName sourcePkgDb
   >=> getVersion
   >=> getLicense
   >=> getAuthorInfo
@@ -131,9 +148,9 @@ extendFlags pkgIx =
   >=> getCategory
   >=> getExtraSourceFiles
   >=> getLibOrExec
+  >=> getSrcDir
   >=> getLanguage
   >=> getGenComments
-  >=> getSrcDir
   >=> getModulesBuildToolsAndDeps pkgIx
 
 -- | Combine two actions which may return a value, preferring the first. That
@@ -151,17 +168,35 @@ maybeToFlag :: Maybe a -> Flag a
 maybeToFlag = maybe NoFlag Flag
 
 -- | Get the package name: use the package directory (supplied, or the current
---   directory by default) as a guess.
-getPackageName :: InitFlags -> IO InitFlags
-getPackageName flags = do
+--   directory by default) as a guess. It looks at the SourcePackageDb to avoid
+--   using an existing package name.
+getPackageName :: SourcePackageDb -> InitFlags -> IO InitFlags
+getPackageName sourcePkgDb flags = do
   guess    <-     traverse guessPackageName (flagToMaybe $ packageDir flags)
               ?>> Just `fmap` (getCurrentDirectory >>= guessPackageName)
 
-  pkgName' <-     return (flagToMaybe $ packageName flags)
-              ?>> maybePrompt flags (prompt "Package name" guess)
-              ?>> return guess
+  let guess' | isPkgRegistered guess = Nothing
+             | otherwise = guess
 
-  return $ flags { packageName = maybeToFlag pkgName' }
+  pkgName' <-     return (flagToMaybe $ packageName flags)
+              ?>> maybePrompt flags (prompt "Package name" guess')
+              ?>> return guess'
+
+  chooseAgain <- if isPkgRegistered pkgName'
+                    then promptYesNo promptOtherNameMsg (Just True)
+                    else return False
+
+  if chooseAgain
+    then getPackageName sourcePkgDb flags
+    else return $ flags { packageName = maybeToFlag pkgName' }
+
+  where
+    isPkgRegistered (Just pkg) = elemByPackageName (packageIndex sourcePkgDb) pkg
+    isPkgRegistered Nothing    = False
+
+    promptOtherNameMsg = "This package name is already used by another " ++
+                         "package on hackage. Do you want to choose a " ++
+                         "different name"
 
 -- | Package version: use 0.1.0.0 as a last resort, but try prompting the user
 --  if possible.
@@ -177,9 +212,9 @@ getVersion flags = do
 getLicense :: InitFlags -> IO InitFlags
 getLicense flags = do
   lic <-     return (flagToMaybe $ license flags)
-         ?>> fmap (fmap (either UnknownLicense id) . join)
+         ?>> fmap (fmap (either UnknownLicense id))
                   (maybePrompt flags
-                    (promptListOptional "Please choose a license" listedLicenses))
+                    (promptList "Please choose a license" listedLicenses (Just BSD3) display True))
   return $ flags { license = maybeToFlag lic }
   where
     listedLicenses =
@@ -245,6 +280,9 @@ getExtraSourceFiles flags = do
 
   return $ flags { extraSrc = extraSrcFiles }
 
+defaultChangeLog :: FilePath
+defaultChangeLog = "ChangeLog.md"
+
 -- | Try to guess things to include in the extra-source-files field.
 --   For now, we just look for things in the root directory named
 --   'readme', 'changes', or 'changelog', with any sort of
@@ -254,12 +292,16 @@ guessExtraSourceFiles flags = do
   dir <-
     maybe getCurrentDirectory return . flagToMaybe $ packageDir flags
   files <- getDirectoryContents dir
-  return $ filter isExtra files
+  let extraFiles = filter isExtra files
+  if any isLikeChangeLog extraFiles
+    then return extraFiles
+    else return (defaultChangeLog : extraFiles)
 
   where
-    isExtra = (`elem` ["README", "CHANGES", "CHANGELOG"])
-            . map toUpper
-            . takeBaseName
+    isExtra = likeFileNameBase ("README" : changeLogLikeBases)
+    isLikeChangeLog = likeFileNameBase changeLogLikeBases
+    likeFileNameBase candidates = (`elem` candidates) . map toUpper . takeBaseName
+    changeLogLikeBases = ["CHANGES", "CHANGELOG"]
 
 -- | Ask whether the project builds a library or executable.
 getLibOrExec :: InitFlags -> IO InitFlags
@@ -285,7 +327,7 @@ getMainFile flags =
   return (flagToMaybe $ mainIs flags)
   ?>> do
     candidates <- guessMainFileCandidates flags
-    let showCandidate = either (++" (does not yet exist)") id
+    let showCandidate = either (++" (does not yet exist, but will be created)") id
         defaultFile = listToMaybe candidates
     maybePrompt flags (either id (either id id) `fmap`
                        promptList "What is the main module of the executable"
@@ -314,14 +356,14 @@ getGenComments flags = do
                  ?>> return (Just False)
   return $ flags { noComments = maybeToFlag (fmap not genComments) }
   where
-    promptMsg = "Include documentation on what each field means (y/n)"
+    promptMsg = "Add informative comments to each field in the cabal file (y/n)"
 
 -- | Ask for the source root directory.
 getSrcDir :: InitFlags -> IO InitFlags
 getSrcDir flags = do
   srcDirs <- return (sourceDirs flags)
              ?>> fmap (:[]) `fmap` guessSourceDir flags
-             ?>> fmap (fmap ((:[]) . either id id) . join) (maybePrompt
+             ?>> fmap (>>= fmap ((:[]) . either id id)) (maybePrompt
                       flags
                       (promptListOptional' "Source directory" ["src"] id))
 
@@ -343,7 +385,7 @@ getModulesBuildToolsAndDeps :: InstalledPackageIndex -> InitFlags -> IO InitFlag
 getModulesBuildToolsAndDeps pkgIx flags = do
   dir <- maybe getCurrentDirectory return . flagToMaybe $ packageDir flags
 
-  -- XXX really should use guessed source roots.
+  -- TODO: really should use guessed source roots.
   sourceFiles <- scanForModules dir
 
   Just mods <-      return (exposedModules flags)
@@ -427,12 +469,17 @@ chooseDep flags (m, Just ps)
       return $ P.Dependency (P.pkgName . head $ pids)
                             (pvpize . maximum . map P.pkgVersion $ pids)
 
-    pvpize :: Version -> VersionRange
-    pvpize v = orLaterVersion v'
-               `intersectVersionRanges`
-               earlierVersion (incVersion 1 v')
-      where v' = (v { versionBranch = take 2 (versionBranch v) })
+-- | Given a version, return an API-compatible (according to PVP) version range.
+--
+-- Example: @0.4.1@ produces the version range @>= 0.4 && < 0.5@ (which is the
+-- same as @0.4.*@).
+pvpize :: Version -> VersionRange
+pvpize v = orLaterVersion v'
+           `intersectVersionRanges`
+           earlierVersion (incVersion 1 v')
+  where v' = (v { versionBranch = take 2 (versionBranch v) })
 
+-- | Increment the nth version component (counting from 0).
 incVersion :: Int -> Version -> Version
 incVersion n (Version vlist tags) = Version (incVersion' n vlist) tags
   where
@@ -629,6 +676,23 @@ writeSetupFile flags = do
     , "main = defaultMain"
     ]
 
+writeChangeLog :: InitFlags -> IO ()
+writeChangeLog flags = when (any (== defaultChangeLog) $ maybe [] id (extraSrc flags)) $ do
+  message flags ("Generating "++ defaultChangeLog ++"...")
+  writeFileSafe flags defaultChangeLog changeLog
+ where
+  changeLog = unlines
+    [ "# Revision history for " ++ pname
+    , ""
+    , "## " ++ pver ++ "  -- YYYY-mm-dd"
+    , ""
+    , "* First version. Released on an unsuspecting world."
+    ]
+  pname = maybe "" display $ flagToMaybe $ packageName flags
+  pver = maybe "" display $ flagToMaybe $ version flags
+
+
+
 writeCabalFile :: InitFlags -> IO Bool
 writeCabalFile flags@(InitFlags{packageName = NoFlag}) = do
   message flags "Error: no package name provided."
@@ -652,6 +716,38 @@ createSourceDirectories flags = case sourceDirs flags of
                                   Just dirs -> forM_ dirs (createDirectoryIfMissing True)
                                   Nothing   -> return ()
 
+-- | Create Main.hs, but only if we are init'ing an executable and
+--   the mainIs flag has been provided.
+createMainHs :: InitFlags -> IO ()
+createMainHs flags@InitFlags{ sourceDirs = Just (srcPath:_)
+                            , packageType = Flag Executable
+                            , mainIs = Flag mainFile } =
+  writeMainHs flags (srcPath </> mainFile)
+createMainHs flags@InitFlags{ sourceDirs = _
+                            , packageType = Flag Executable
+                            , mainIs = Flag mainFile } =
+  writeMainHs flags mainFile
+createMainHs _ = return ()
+
+-- | Write a main file if it doesn't already exist.
+writeMainHs :: InitFlags -> FilePath -> IO ()
+writeMainHs flags mainPath = do
+  dir <- maybe getCurrentDirectory return (flagToMaybe $ packageDir flags)
+  let mainFullPath = dir </> mainPath
+  exists <- doesFileExist mainFullPath
+  unless exists $ do
+      message flags $ "Generating " ++ mainPath ++ "..."
+      writeFileSafe flags mainFullPath mainHs
+
+-- | Default Main.hs file.  Used when no Main.hs exists.
+mainHs :: String
+mainHs = unlines
+  [ "module Main where"
+  , ""
+  , "main :: IO ()"
+  , "main = putStrLn \"Hello, Haskell!\""
+  ]
+
 -- | Move an existing file, if there is one, and the overwrite flag is
 --   not set.
 moveExistingFile :: InitFlags -> FilePath -> IO ()
@@ -682,6 +778,7 @@ findNewName oldName = findNewName' 0
 --   pretty-printing code to generate the file.
 generateCabalFile :: String -> InitFlags -> String
 generateCabalFile fileName c =
+  (++ "\n") .
   renderStyle style { lineLength = 79, ribbonsPerLine = 1.1 } $
   (if minimal c /= Flag True
     then showComment (Just $ "Initial " ++ fileName ++ " generated by cabal "
@@ -695,7 +792,7 @@ generateCabalFile fileName c =
                 True
 
        , field  "version"       (version       c)
-                (Just $ "The package version.  See the Haskell package versioning policy (PVP) for standards guiding when and how versions should be incremented.\nhttp://www.haskell.org/haskellwiki/Package_versioning_policy\n"
+                (Just $ "The package version.  See the Haskell package versioning policy (PVP) for standards guiding when and how versions should be incremented.\nhttps://wiki.haskell.org/Package_versioning_policy\n"
                 ++ "PVP summary:      +-+------- breaking API changes\n"
                 ++ "                  | | +----- non-breaking API additions\n"
                 ++ "                  | | | +--- code changes with no API change")
@@ -721,9 +818,11 @@ generateCabalFile fileName c =
                 (Just "The license under which the package is released.")
                 True
 
-       , fieldS "license-file" (Flag "LICENSE")
-                (Just "The file containing the license text.")
-                True
+       , case (license c) of
+           Flag PublicDomain -> empty
+           _ -> fieldS "license-file" (Flag "LICENSE")
+                       (Just "The file containing the license text.")
+                       True
 
        , fieldS "author"        (author       c)
                 (Just "The package author(s).")
@@ -733,9 +832,11 @@ generateCabalFile fileName c =
                 (Just "An email address to which users can send suggestions, bug reports, and patches.")
                 True
 
-       , fieldS "copyright"     NoFlag
-                (Just "A copyright notice.")
-                True
+       , case (license c) of
+           Flag PublicDomain -> empty
+           _ -> fieldS "copyright"     NoFlag
+                       (Just "A copyright notice.")
+                       True
 
        , fieldS "category"      (either id display `fmap` category c)
                 Nothing
