@@ -1,4 +1,4 @@
-#if defined(__GLASGOW_HASKELL__) && (__GLASGOW_HASKELL__ >= 612)
+#if defined(__GLASGOW_HASKELL__) && (__GLASGOW_HASKELL__ >= 612) && !defined(CABAL_NO_THREADED)
 /* Since version 6.12, GHC's threaded RTS includes a getNumberOfProcessors
    function, so we try to use that if available. cabal-install is always built
    with -threaded nowadays.  */
@@ -8,7 +8,7 @@
 
 #ifndef HAS_GET_NUMBER_OF_PROCESSORS
 
-#ifdef _WIN32
+#if defined(_WIN32) && !defined(__CYGWIN__)
 #include <windows.h>
 #elif MACOS
 #include <sys/param.h>
@@ -18,7 +18,7 @@
 #endif
 
 int getNumberOfProcessors() {
-#ifdef WIN32
+#if defined(_WIN32) && !defined(__CYGWIN__)
     SYSTEM_INFO sysinfo;
     GetSystemInfo(&sysinfo);
     return sysinfo.dwNumberOfProcessors;
